package com.tzld.piaoquan.recommend.server.service.rank;


import com.typesafe.config.Config;
import com.tzld.piaoquan.recommend.server.common.base.RankItem;
import com.tzld.piaoquan.recommend.server.common.base.UserFeature;
import com.tzld.piaoquan.recommend.server.common.enums.AppTypeEnum;
import com.tzld.piaoquan.recommend.server.gen.recommend.RecommendRequest;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.flowpool.FlowPoolConstants;
import com.tzld.piaoquan.recommend.server.service.recall.RecallResult;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.*;
import com.tzld.piaoquan.recommend.server.service.score.ScorerPipeline;
import com.tzld.piaoquan.recommend.server.service.score.ScorerUtils;
import com.tzld.piaoquan.recommend.server.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.RandomUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;

import java.util.*;
import java.util.stream.Collectors;

/**
 * @author dyp
 */
@Service
@Slf4j
public class RankService {
    @Autowired
    private RedisTemplate<String, String> redisTemplate;


    // LR scores
    public RankResult rank(RecommendRequest recommendRequest,
                           RankParam param,
                           RecallResult recallResult,
                           UserFeature userFeature,
                           List<RankItem> videoRankFeatures,
                           Config config) {
        if (param == null
                || param.getRecallResult() == null
                || CollectionUtils.isEmpty(param.getRecallResult().getData())) {
            return null;
        }

        if (param.isSpecialRecommend()) {
            Optional<RecallResult.RecallData> data = param.getRecallResult().getData().stream()
                    .filter(d -> d.getPushFrom().equals(SpecialRecallStrategy.PUSH_FROM))
                    .findFirst();
            if (data.isPresent()
                    && data.get() != null) {
                return new RankResult(data.get().getVideos());
            }
            return null;
        }

        ScorerPipeline scorerPipeline = ScorerUtils.getScorerPipeline(config);
        // TODO  merge 后返回待排序的video list
        List<Video> recallVideos = recallResult.mergeRecallVideos();
        // TODO 转化成rankitem
        // convert List<Video> to List<RankItem>
        List<RankItem> rankItem = new ArrayList<RankItem>();

        List<RankItem> rovRecallRank  =
                scorerPipeline.scoring(recommendRequest, param, userFeature, rankItem);

        log.info("mergeAndRankRovRecall rovRecallRank={}", JSONUtils.toJson(rovRecallRank));
        List<Video> flowPoolRank = mergeAndRankFlowPoolRecall(param);
        log.info("mergeAndRankFlowPoolRecall flowPoolRank={}", JSONUtils.toJson(flowPoolRank));

        removeDuplicate(param, rovRecallRank, flowPoolRank);
        // convert List<RankItem> to List<Video>



        log.info("removeDuplicate rovRecallRank={}, flowPoolRank={}",
                JSONUtils.toJson(rovRecallRank),
                JSONUtils.toJson(flowPoolRank));

        // 融合排序
        return mergeAndSort(param, rovRecallRank, flowPoolRank);


    }


    public RankResult rank(RankParam param) {


        if (param == null
                || param.getRecallResult() == null
                || CollectionUtils.isEmpty(param.getRecallResult().getData())) {
            return null;
        }

        if (param.isSpecialRecommend()) {
            Optional<RecallResult.RecallData> data = param.getRecallResult().getData().stream()
                    .filter(d -> d.getPushFrom().equals(SpecialRecallStrategy.PUSH_FROM))
                    .findFirst();
            if (data.isPresent()
                    && data.get() != null) {
                return new RankResult(data.get().getVideos());
            }
            return null;
        }

        List<Video> rovRecallRank = mergeAndRankRovRecall(param);
        log.info("mergeAndRankRovRecall rovRecallRank={}", JSONUtils.toJson(rovRecallRank));
        List<Video> flowPoolRank = mergeAndRankFlowPoolRecall(param);
        log.info("mergeAndRankFlowPoolRecall flowPoolRank={}", JSONUtils.toJson(flowPoolRank));

        removeDuplicate(param, rovRecallRank, flowPoolRank);

        log.info("removeDuplicate rovRecallRank={}, flowPoolRank={}",
                JSONUtils.toJson(rovRecallRank),
                JSONUtils.toJson(flowPoolRank));

        // 融合排序
        return mergeAndSort(param, rovRecallRank, flowPoolRank);
    }

    private List<Video> mergeAndRankRovRecall(RankParam param) {
        List<Video> rovRecallRank = new ArrayList<>();
        rovRecallRank.addAll(extractAndSort(param, RegionHRecallStrategy.PUSH_FORM));
        rovRecallRank.addAll(extractAndSort(param, RegionHDupRecallStrategy.PUSH_FORM));
        rovRecallRank.addAll(extractAndSort(param, Region24HRecallStrategy.PUSH_FORM));
        rovRecallRank.addAll(extractAndSort(param, RegionRelative24HRecallStrategy.PUSH_FORM));
        rovRecallRank.addAll(extractAndSort(param, RegionRelative24HDupRecallStrategy.PUSH_FORM));
        // TODO 为什么sim recall 和 return recall 不去重
        rovRecallRank.addAll(extractAndSort(param, SimHotVideoRecallStrategy.PUSH_FORM));
        rovRecallRank.addAll(extractAndSort(param, ReturnVideoRecallStrategy.PUSH_FORM));

        // 去重
        if (CollectionUtils.isNotEmpty(rovRecallRank)) {
            Set<Long> videoIds = new HashSet<>();
            Iterator<Video> ite = rovRecallRank.iterator();
            while (ite.hasNext()) {
                Video v = ite.next();
                if (videoIds.contains(v.getVideoId())) {
                    ite.remove();
                    continue;
                }
                videoIds.add(v.getVideoId());
            }
        }

        // 排序实验
        if (param.getAbCode().equals("60054")
                || param.getAbCode().equals("60068")
                || param.getAbCode().equals("60081")
                || param.getAbCode().equals("60084")
                || param.getAbCode().equals("60087")
                || param.getAbCode().equals("60088")
                || param.getAbCode().equals("60089")
                || param.getAbCode().equals("60090")
                || param.getAbCode().equals("60091")
                || param.getAbCode().equals("60092")
                || param.getAbCode().equals("60093")
                || param.getAbCode().equals("60094")
                || param.getAbCode().equals("60095")
                || param.getAbCode().equals("60096")) {

            List<String> videoIdKeys = rovRecallRank.stream()
                    .map(t -> param.getRankKeyPrefix() + t.getVideoId())
                    .collect(Collectors.toList());
            List<String> videoScores = redisTemplate.opsForValue().multiGet(videoIdKeys);
            log.info("rank mergeAndRankRovRecall videoIdKeys={}, videoScores={}", JSONUtils.toJson(videoIdKeys),
                    JSONUtils.toJson(videoScores));
            if (CollectionUtils.isNotEmpty(videoScores)
                    && videoScores.size() == rovRecallRank.size()) {
                for (int i = 0; i < videoScores.size(); i++) {
                    rovRecallRank.get(i).setSortScore(NumberUtils.toDouble(videoScores.get(i), 0.0));
                }
                Collections.sort(rovRecallRank, Comparator.comparingDouble(o -> -o.getSortScore()));
            }
        }
        return rovRecallRank;
    }

    private List<Video> mergeAndRankFlowPoolRecall(RankParam param) {
        if (param.getAppType() == AppTypeEnum.LAO_HAO_KAN_VIDEO.getCode()
                || param.getAppType() == AppTypeEnum.ZUI_JING_QI.getCode()) {
            if (param.getAbCode().equals("60054")
                    || param.getAbCode().equals("60068")
                    || param.getAbCode().equals("60081")
                    || param.getAbCode().equals("60084")) {
                return extractAndSort(param, FlowPoolConstants.QUICK_PUSH_FORM);
            } else {
                return Collections.emptyList();
            }
        } else {
            List<Video> quickFlowPoolVideos = extractAndSort(param, FlowPoolConstants.QUICK_PUSH_FORM);
            if (CollectionUtils.isNotEmpty(quickFlowPoolVideos)) {
                return quickFlowPoolVideos;
            } else {
                return extractAndSort(param, FlowPoolConstants.PUSH_FORM);
            }
        }
    }

    private List<Video> extractAndSort(RankParam param, String pushFrom) {
        if (param == null
                || param.getRecallResult() == null
                || CollectionUtils.isEmpty(param.getRecallResult().getData())) {
            return Collections.emptyList();
        }
        Optional<RecallResult.RecallData> data = param.getRecallResult().getData().stream()
                .filter(d -> d.getPushFrom().equals(pushFrom))
                .findFirst();
        if (data.isPresent()
                && data.get() != null
                && CollectionUtils.isNotEmpty(data.get().getVideos())) {
            List<Video> result = data.get().getVideos();
            Collections.sort(result, Comparator.comparingDouble(o -> -o.getRovScore()));
            return result;
        }
        return Collections.emptyList();
    }

    private double getFlowPoolP(RankParam param) {
        if (param.getAppType() == AppTypeEnum.LAO_HAO_KAN_VIDEO.getCode()
                || param.getAppType() == AppTypeEnum.ZUI_JING_QI.getCode()) {
            return param.getFlowPoolP();
        } else {
            List<Video> quickFlowPoolVideos = extractAndSort(param, FlowPoolConstants.QUICK_PUSH_FORM);
            if (CollectionUtils.isNotEmpty(quickFlowPoolVideos)) {
                String quick_flow_pool_P = redisTemplate.opsForValue().get("flow:pool:quick:distribute:rate:3");
                return NumberUtils.toDouble(quick_flow_pool_P, 0);
            } else {
                return param.getFlowPoolP();
            }
        }
    }

    private void removeDuplicate(RankParam param, List<Video> rovRecallRank, List<Video> flowPoolRank) {
        // TODO 重构 rov和流量池 融合排序
        //    去重原则：
        //        如果视频在ROV召回池topK，则保留ROV召回池，否则保留流量池
        // 1 rov pool topK
        Set<Long> rovTopKVideoIds = new HashSet<>();
        for (int i = 0; i < param.getTopK() && i < rovRecallRank.size(); i++) {
            rovTopKVideoIds.add(rovRecallRank.get(i).getVideoId());
        }

        // 2 flow pool 移除topK视频
        Set<Long> flowPoolVideoIds = new HashSet<>();
        Iterator<Video> flowRecallRankIte = flowPoolRank.iterator();
        while (flowRecallRankIte.hasNext()) {
            Video data = flowRecallRankIte.next();
            if (rovTopKVideoIds.contains(data.getVideoId())) {
                flowRecallRankIte.remove();
            } else {
                flowPoolVideoIds.add(data.getVideoId());
            }
        }

        // rov pool 移除flow中的视频
        Iterator<Video> rovRecallRankIte = rovRecallRank.iterator();
        while (rovRecallRankIte.hasNext()) {
            Video data = rovRecallRankIte.next();
            if (flowPoolVideoIds.contains(data.getVideoId())) {
                rovRecallRankIte.remove();
            }
        }
    }

    private RankResult mergeAndSort(RankParam param, List<Video> rovRecallRank, List<Video> flowPoolRank) {
        if (CollectionUtils.isEmpty(rovRecallRank)) {
            if (param.getSize() < flowPoolRank.size()) {
                return new RankResult(flowPoolRank.subList(0, param.getSize()));
            } else {
                return new RankResult(flowPoolRank);
            }
        }

        List<Video> result = new ArrayList<>();
        for (int i = 0; i < param.getTopK() && i < rovRecallRank.size(); i++) {
            result.add(rovRecallRank.get(i));
        }

        double flowPoolP = getFlowPoolP(param);
        int flowPoolIndex = 0;
        int rovPoolIndex = param.getTopK();

        for (int i = 0; i < param.getSize() - param.getTopK(); i++) {
            double rand = RandomUtils.nextDouble(0, 1);
            log.info("rand={}, flowPoolP={}", rand, flowPoolP);
            if (rand < flowPoolP) {
                if (flowPoolIndex < flowPoolRank.size()) {
                    result.add(flowPoolRank.get(flowPoolIndex++));
                } else {
                    break;
                }
            } else {
                if (rovPoolIndex < rovRecallRank.size()) {
                    result.add(rovRecallRank.get(rovPoolIndex++));
                } else {
                    break;
                }
            }
        }
        if (rovPoolIndex >= rovRecallRank.size()) {
            for (int i = flowPoolIndex; i < flowPoolRank.size() && result.size() < param.getSize(); i++) {
                result.add(flowPoolRank.get(i));
            }
        }
        if (flowPoolIndex >= flowPoolRank.size()) {
            for (int i = rovPoolIndex; i < rovRecallRank.size() && result.size() < param.getSize(); i++) {
                result.add(rovRecallRank.get(i));
            }
        }
        return new RankResult(result);
    }

}
