package com.tzld.piaoquan.recommend.server.service.score;

import com.tzld.piaoquan.recommend.server.common.base.RankItem;
import com.tzld.piaoquan.recommend.server.common.base.UserFeature;
import com.tzld.piaoquan.recommend.server.gen.recommend.RecommendRequest;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.rank.RankParam;
import com.tzld.piaoquan.recommend.server.service.rank.RankResult;
import com.tzld.piaoquan.recommend.server.service.recall.RecallResult;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;


public class ScorerPipeline {
    public static final int corePoolSize = 128;
    public static final int SCORE_TIME_OUT = 400;
    public static final Logger LOGGER = LoggerFactory.getLogger(ScorerPipeline.class);
    public static final ExecutorService executorService = Executors.newFixedThreadPool(corePoolSize);

    public List<AbstractScorer> scorers;

    public ScorerPipeline(List<AbstractScorer> scorers) {
        this.scorers = scorers;
    }

    /**
     * scoring
     * @return
     */
    public List<RankItem> scoring(final RecommendRequest recommendRequest,
                               final RankParam param,
                               final UserFeature userFeature,
                               final List<RankItem> rankItems) {
        // check recall is empty?
        if (CollectionUtils.isEmpty(param.getRecallResult().getData())) {
            // log.error
        }

        if (CollectionUtils.isEmpty(scorers)) {
            // log.error()
        }
        long scoreStart = System.currentTimeMillis();
        for (final AbstractScorer scorer : scorers) {
            if (!scorer.isEnable()) {
                continue;
            }

            final int beforeSize = rankItems.size();
            final long startTime = System.currentTimeMillis();

            String fullScorerName = scorer.getScorerConfigInfo().getScorerName();
            String[] scorerNames = fullScorerName.split("\\.");
            final String scorerName = scorerNames.length > 0 ? scorerNames[scorerNames.length - 1] : fullScorerName;

            final List<RankItem> scoreRankerItems = rankItems;
            Callable<List<RankItem>> callable = new Callable<List<Video>>() {
                @Override
                public List<RankItem> call() throws Exception {
                    return scorer.scoring(recommendRequest, param, userFeature, rankItems);
                }
            };
           //
            long spentTime = System.currentTimeMillis() - startTime;
            LOGGER.debug("after scorer [{}], spentTime [{}], before size [{}], remaining size [{}]",
                    new Object[]{scorerName, spentTime, beforeSize, items.size()});
        }
        return items;
    }
}
